/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Odonata.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.xml.xqueryevaluator.eval;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.Reader;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParserFactory;

import org.w3c.dom.Element;
import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;
import org.xquark.datastructure.BufferTuple;
import org.xquark.datastructure.EvaluationVisitor;
import org.xquark.mediator.plan.DynamicContext;
import org.xquark.mediator.runtime.MediatorException;
import org.xquark.schema.SchemaManager;
import org.xquark.schema.validation.SchemaValidationContext;
import org.xquark.schema.validation.ValidatingSchemaFilter;
import org.xquark.schema.validation.ValidatingSchemaHandler;
import org.xquark.util.DOMReader;
import org.xquark.xpath.datamodel.TypedNode;
import org.xquark.xquery.parser.FLWRExpression;
import org.xquark.xquery.parser.LocatedExpression;
import org.xquark.xquery.parser.Step;
import org.xquark.xquery.parser.ValueString;
import org.xquark.xquery.parser.Variable;
import org.xquark.xquery.parser.XQueryException;
import org.xquark.xquery.parser.XQueryExpression;
import org.xquark.xquery.parser.primitivefunctions.fnfunctions.FunctionDOC;

/**
 *  
 */
public class EvalFilter {

    // **********************************************************************
    // * VERSIONING
    // **********************************************************************

    private static final String RCSRevision = "$Revision: 1.14 $";

    private static final String RCSName = "$Name: XQ1_1_1 $";

    // **********************************************************************
    // * CLASS VARIABLES
    // **********************************************************************

    private static SAXParserFactory saxParserFactory = null;
    static {
        // Create a JAXP SAXParserFactory and configure it
        saxParserFactory = SAXParserFactory.newInstance();
        saxParserFactory.setNamespaceAware(true);
        saxParserFactory.setValidating(false);
    }

    private XMLReader saxReader = null;

    private InputSource inputSource = null;

    private XPPReader xppReader = null;

    private BufferTuple buftuples = null;

    private EvaluationVisitor evalVisitor = null;
    
    private ArrayList nodes = null;
    private DOMReader domReader = null;

    // **********************************************************************
    // * INITIALIZATION
    // **********************************************************************

    public EvalFilter(BufferTuple buftuples, DynamicContext context, ArrayList paths, XQueryExpression expr) throws MediatorException {
        this(buftuples, context, paths, expr, true);
    }

    public EvalFilter(BufferTuple buftuples, DynamicContext context, ArrayList paths, XQueryExpression expr, boolean useXPP) throws MediatorException {
        this.buftuples = buftuples;
        try {
            SchemaValidationContext svc = new SchemaValidationContext(new SchemaManager(context.getSchemaManager()));
            if (expr instanceof Variable) {

                SAX2TypedDOM handler = new SAX2TypedDOM((Variable) expr, buftuples, context, paths, svc);

                if (evalVisitor == null)
                    evalVisitor = new EvaluationVisitor(context.getSchemaManager(), context.getCurrentTuples());
                else
                    evalVisitor.reset(context.getCurrentTuples());
                evalVisitor.setReturnType(EvaluationVisitor.NODE_TYPE);
                try {
                    expr.accept(evalVisitor);
                } catch (XQueryException xqe) {
                    throw new MediatorException("Could not find external variable in current context : " + expr.getStringValue());
                }
                nodes = evalVisitor.getResNodes();
                //TypedNode node = (TypedNode) nodes.get(0);
                //for (int i=0;i<nodes.size();i++) {}

                domReader = new DOMReader();
                // should be done but better not so everything is consistent with Document
                //reader.setFeature(DOMReader.FRAGMENT_FEATURE, true);
                
                ValidatingSchemaFilter filter = new ValidatingSchemaFilter(domReader, svc, false);
                filter.setContentHandler(handler);
                filter.setLexicalHandler(handler);
                filter.setErrorHandler(new ErrorHandler() {
                    public void warning(SAXParseException ex) throws SAXException {
                    }

                    public void error(SAXParseException ex) throws SAXException {
                        throw ex;
                    }

                    public void fatalError(SAXParseException ex) throws SAXException {
                        throw ex;
                    }
                });
                saxReader = filter;
            } else if (expr instanceof FLWRExpression) {
                FLWRExpression flwr = (FLWRExpression) expr;

                SAX2TypedDOM handler = new SAX2TypedDOM((FLWRExpression) flwr, buftuples, context, paths, svc);
                Variable var = (Variable) flwr.getVariables().get(0);
                String doc = null;
                XQueryExpression arg = null;
                if (var.getExpression() instanceof LocatedExpression)
                    arg = ((FunctionDOC) ((Step) var.getExpression().getSteps().get(0)).getExpression()).getArgument(0);
                else if (var.getExpression() instanceof FunctionDOC)
                    arg = ((FunctionDOC) var.getExpression()).getArgument(0);

                if (arg instanceof ValueString)
                    doc = ((ValueString) arg).getStringValue();
                else {
                    if (evalVisitor == null)
                        evalVisitor = new EvaluationVisitor(context.getSchemaManager(), context.getCurrentTuples());
                    else
                        evalVisitor.reset(context.getCurrentTuples());
                    evalVisitor.setReturnType(EvaluationVisitor.VALUE_TYPE);
                    try {
                        arg.accept(evalVisitor);
                    } catch (XQueryException xqe) {
                        throw new MediatorException("Incorrect argument of function doc : " + arg.getStringValue());
                    }
                    Comparable comp = evalVisitor.getResValue();
                    if (comp != null)
                        doc = comp.toString();
                }
                if (doc == null)
                    throw new MediatorException("Incorrect argument of function doc : " + arg.getStringValue());
                if (doc.startsWith("*:"))
                    doc = doc.substring(2);
                URL url;
                if (context.getBaseURI() != null)
                    url = new URL(new URL(context.getBaseURI()), doc);
                else
                    url = new URL(doc);

                if (useXPP) {
                    xppReader = new XPPReader();
                    ValidatingSchemaHandler filter = new ValidatingSchemaHandler(svc, false);
                    filter.setContentHandler(handler);
                    filter.setLexicalHandler(handler);
                    filter.setErrorHandler(new ErrorHandler() {
                        public void warning(SAXParseException ex) throws SAXException {
                        }

                        public void error(SAXParseException ex) throws SAXException {
                            throw ex;
                        }

                        public void fatalError(SAXParseException ex) throws SAXException {
                            throw ex;
                        }
                    });
                    xppReader.setContentHandler(filter);
                    xppReader.setLexicalHandler(filter);
                    xppReader.setErrorHandler(filter);
                    xppReader.setInput(url.toString(), url.openStream());
                } else {
                    XMLReader reader = saxParserFactory.newSAXParser().getXMLReader();
                    ValidatingSchemaFilter filter = new ValidatingSchemaFilter(reader, svc, false);
                    filter.setContentHandler(handler);
                    filter.setLexicalHandler(handler);
                    filter.setErrorHandler(new ErrorHandler() {
                        public void warning(SAXParseException ex) throws SAXException {
                        }

                        public void error(SAXParseException ex) throws SAXException {
                            throw ex;
                        }

                        public void fatalError(SAXParseException ex) throws SAXException {
                            throw ex;
                        }
                    });
                    saxReader = filter;
                    inputSource = new InputSource(url.toString());
                }
            } else { // error
                throw new MediatorException("ERROR incorrect expression type for evalFilter");
            }
        } catch (SAXException e) {
            throw new MediatorException(e.getMessage());
        } catch (MalformedURLException e) {
            throw new MediatorException(e.getMessage());
        } catch (FileNotFoundException e) {
            throw new MediatorException("Could not find file : " + e.getMessage());
        } catch (IOException e) {
            throw new MediatorException(e.getMessage());
        } catch (ParserConfigurationException e) {
            throw new MediatorException(e.getMessage());
        }
    }

    // **********************************************************************
    // * ACCESS
    // **********************************************************************

    public boolean evaluate() throws MediatorException {
        if (xppReader != null)
            return evaluateXPP();
        else
            return evaluateSAX();
    }

    private boolean evaluateSAX() throws MediatorException {
        try {
            if (inputSource != null) {
                saxReader.parse(inputSource);
                return true;
            } else { 
                if (nodes == null || nodes.isEmpty()) return true;
                for (int i=0;i<nodes.size();i++) {
                    domReader.setRoot((Element)nodes.get(i));
                    saxReader.parse("");
                }
                return true;
            }
            //return false;
        } catch (IOException e) {
            throw new MediatorException(e.getMessage());
        } catch (SAXException e) {
            throw new MediatorException(e.getMessage());
        }
    }

    private boolean evaluateXPP() throws MediatorException {
        try {
            boolean isDone = false;
            if (xppReader.isAtStart())
                xppReader.getContentHandler().startDocument();
            while (buftuples.isEmpty() && !isDone) {
                isDone = xppReader.parseNextElement();
            }
            if (isDone)
                xppReader.getContentHandler().endDocument();
            return isDone;
        } catch (IOException e) {
            throw new MediatorException(e.getMessage());
        } catch (SAXException e) {
            throw new MediatorException(e.getMessage());
        }
    }

}